#!/usr/bin/python3
#
# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
import contextlib
import json
import logging
import os
import re
import subprocess
import sys
import time
from argparse import ArgumentParser, ArgumentTypeError, Namespace, RawDescriptionHelpFormatter
from datetime import datetime
from os.path import isfile, join
from typing import Callable, List, Optional, Tuple

import lazy_object_proxy
from ldap3 import Entry

from ucsschool_id_connector.ldap_access import LDAPAccess
from ucsschool_id_connector.models import SchoolAuthorityConfiguration
from ucsschool_id_connector.utils import ConsoleAndFileLogging

try:
    from verify import dn_to_id, ldap_object_to_school, ldap_object_to_school_class

    from .provisioning_api.exceptions import ApiException
except ImportError:

    sys.path.append("/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/plugins/packages")
    from idbroker.provisioning_api.exceptions import ApiException
    from idbroker.verify import dn_to_id, ldap_object_to_school, ldap_object_to_school_class


try:
    # works in IDE, but not on the cli
    from .id_broker_client import (
        IDBrokerError,
        IDBrokerObject,
        IDBrokerSchool,
        IDBrokerSchoolClass,
        School,
        SchoolClass,
    )
except ImportError:
    # works on the cli in ID Connector Docker container
    # make `idbroker` package importable
    import sys

    sys.path.append("/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/plugins/packages")

    from idbroker.id_broker_client import (
        IDBrokerError,
        IDBrokerObject,
        IDBrokerSchool,
        IDBrokerSchoolClass,
        School,
        SchoolClass,
    )

SCHOOL_ATTRS = ["entryUUID", "ou", "displayName"]
SCHOOL_CLASS_ATTRS = ["entryUUID", "cn", "description", "uniqueMember"]
USER_ATTRS = ["entryUUID", "uid", "givenName", "sn", "memberOf", "ucsschoolRole"]
SCHOOL_CLASS_DN_PATTERN = re.compile(
    r"^cn=(?P<cn>.+?),cn=klassen,cn=schueler,cn=groups,ou=(?P<ou>.+?),dc=.+$"
)

MAX_THREADS = 32


logger: logging.Logger = lazy_object_proxy.Proxy(
    lambda: ConsoleAndFileLogging.get_logger(os.path.basename(__file__))
)
ldap = LDAPAccess()

DEFAULT_POOL_SIZE: int = 4
auth_config: Optional[SchoolAuthorityConfiguration] = None
SCHOOL_CLIENT: Optional[IDBrokerSchool] = None
SCHOOLCLASS_CLIENT: Optional[IDBrokerSchoolClass] = None
SYNC_MEMBERS: Optional[bool] = None
DRY_MODE: Optional[bool] = None
DEFAULT_RETRY_SLEEP_TIME: int = 10
TASK_SEM: Optional[asyncio.Semaphore] = None


def list_dir(path: str) -> List:
    return [
        join(os.path.abspath(path), f)
        for f in os.listdir(os.path.abspath(path))
        if isfile(join(os.path.abspath(path), f))
    ]


async def retry(func, *args, max_attempts=3, sleep_time: int = DEFAULT_RETRY_SLEEP_TIME):
    for attempt in range(max_attempts):
        try:
            return await func(*args)
        except IDBrokerError:
            if attempt >= max_attempts - 1:
                logger.error("ERROR in %r %r", func.__name__, args)
                raise
            time.sleep(sleep_time)


async def sync_object(obj: IDBrokerObject) -> Tuple[str, bool, bool, bool]:
    created = False
    updated = False

    if isinstance(obj, School):
        obj_type = School
        client = SCHOOL_CLIENT
    elif isinstance(obj, SchoolClass):
        obj_type = SchoolClass
        if SYNC_MEMBERS:
            if obj.members:
                obj.members = [await dn_to_id(member) for member in obj.members]
        else:
            obj.members = []
        client = SCHOOLCLASS_CLIENT
    else:
        try:
            raise ValueError(f"Could not identify type of given object {obj.name} (type={type(obj)})")
        except Exception as exc:
            raise ValueError(
                f"Could not identify type of given object {obj.user_name} (type={type(obj)})"
            ) from exc

    name = f"{obj.school}-{obj.name}" if obj_type == SchoolClass else obj.name
    logger.info("Checking %s %r...", obj_type.__name__, name)
    exists = await retry(client.exists, obj.id)
    try:
        if exists:
            if obj_type != School and SYNC_MEMBERS:
                logger.info("Updating %s %r...", obj_type.__name__, name)
                if not DRY_MODE:
                    await retry(client.update, obj)
                updated = True
        else:
            logger.info("Creating %s %r...", obj_type.__name__, name)
            if not DRY_MODE:
                await retry(client.create, obj)
            created = True
    except (asyncio.exceptions.TimeoutError, ApiException, IDBrokerError) as exc:
        logger.info("ERROR handling %s %r: %s.", obj_type.__name__, name, str(exc))

    return name, exists, created, updated


async def limited_func(sem: asyncio.Semaphore, func: Callable, *args):
    async with sem:
        return await func(*args)


async def parallelize(funcs, args, max_workers: int = DEFAULT_POOL_SIZE):
    task_limiter = asyncio.Semaphore(max_workers)
    tasks = [limited_func(task_limiter, func, arg) for func in funcs for arg in args]
    return await asyncio.gather(*tasks)


async def get_classes(
    ldap_filter: str,
) -> List[SchoolClass]:
    ldap_filter = f"(&(ucsschoolRole=school_class:school:{ldap_filter})(memberUid=*))"
    logger.info("Retrieving classes with filter %r...", ldap_filter)
    objects: List[Entry] = await ldap.search(filter_s=ldap_filter, attributes=SCHOOL_CLASS_ATTRS)
    return [ldap_object_to_school_class(o) for o in objects]


async def get_ous(
    ldap_filter: str,
) -> List[School]:
    ldap_filter = f"(&(objectClass=ucsschoolOrganizationalUnit)(ou={ldap_filter}))"
    logger.info("Retrieving OUs with filter %r...", ldap_filter)
    objects: List[Entry] = await ldap.search(
        base=os.environ.get("ldap_base"), filter_s=ldap_filter, attributes=SCHOOL_ATTRS
    )
    return [ldap_object_to_school(o) for o in objects]


async def sync_complete(school: str, dry_mode: bool, max_workers: int) -> None:
    ous = await get_ous(school)
    logger.info("Found %d schools.", len(ous))
    groups = await get_classes(school)
    logger.info("Found %d classes.", len(groups))
    t0 = time.time()
    await parallelize(funcs=[sync_object], args=ous, max_workers=1)
    logger.info("Synchronized %d schools in %0.2f minutes.", len(ous), (time.time() - t0) / 60.0)
    t0 = time.time()
    await parallelize(funcs=[sync_object], args=groups, max_workers=max_workers)
    logger.info("Synchronized %d classes in %0.2f minutes.", len(groups), (time.time() - t0) / 60.0)


def setup(config_path: str, opt: Namespace) -> None:
    global auth_config, SYNC_MEMBERS, SCHOOL_CLIENT, SCHOOLCLASS_CLIENT, DRY_MODE

    for f in list_dir(config_path):
        if f.split(".")[1] == "json":
            config = json.load(open(f))
            if "id_broker" in config["plugin_configs"]:
                with contextlib.suppress(Exception):
                    auth_config = SchoolAuthorityConfiguration(
                        name=config["name"],
                        url=config["url"],
                        plugin_configs=config["plugin_configs"],
                    )
                    break
    SCHOOL_CLIENT = IDBrokerSchool(school_authority=auth_config, plugin_name="id_broker")
    SCHOOLCLASS_CLIENT = IDBrokerSchoolClass(school_authority=auth_config, plugin_name="id_broker")
    SYNC_MEMBERS = opt.members
    DRY_MODE = opt.dry_mode


def num_task_range():
    def checker(arg):
        if arg.isdigit():
            tasks = int(arg)
        else:
            raise ArgumentTypeError(
                "Invalid argument type! The entered value is not type int or negative."
            )
        if tasks < 1:
            raise ArgumentTypeError("At least one task is needed to run.")
        if tasks > MAX_THREADS:
            raise ArgumentTypeError(
                f"The maximum allowed value for concurrent tasks is {MAX_THREADS} (Entered: {arg})"
            )
        return tasks

    return checker


if __name__ == "__main__":

    parser = ArgumentParser(description=__doc__, formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(
        "-d",
        "--dry-mode",
        help="fakes complete sync with no commitment to the IDBroker and displays which objects would be created",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--members",
        help="enables the sync of users",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--school",
        help="only sync a single school with its classes (and optionally members if option -m is used)",
        type=str,
    )
    parser.add_argument(
        "-p",
        "--parallelism",
        type=num_task_range(),
        default=DEFAULT_POOL_SIZE,
        help=f"defines the number of parallel tasks that are used for the specified task (min=1, "
        f"max={MAX_THREADS})",
    )
    parser.add_argument(
        "-c",
        "--config_path",
        default="/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/school_authorities",
        help="defines the path where the configuration for school_authorities can be found",
    )

    parser.epilog = """Examples:
    $ initial_sync.py
        syncs all ous and classes without members (add -m to include members)

    $ initial_sync.py -s OU:str
        syncs all classes of the ou specified by OU without members (add -m to include members)

    $ initial_sync.py -d
        fake sync as preview for all ous and classes without members (add -m to include members)

    $ initial_sync.py -d -s OU:str
        fake sync of singular ou specified by OU without members (add -m to include members)

    """

    opt = parser.parse_args()

    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(ConsoleAndFileLogging.get_formatter())
    logger.addHandler(handler)
    logger.info("STARTING improved first time sync with arguments %s...", opt)
    logger.info("Logging to '/var/log/univention/ucsschool-id-connector/queues.log'.")
    start = datetime.now()
    logger.info("Running with %d tasks.", int(opt.parallelism))

    setup(config_path=opt.config_path, opt=opt)
    school = str(opt.school) if opt.school else "*"
    try:
        logger.info("Stopping UCS@school ID Connector...")
        subprocess.check_output(["/etc/init.d/ucsschool-id-connector", "stop"])
        asyncio.run(
            sync_complete(school=school, dry_mode=opt.dry_mode, max_workers=int(opt.parallelism))
        )
    finally:
        logger.info("Starting UCS@school ID Connector...")
        subprocess.check_output(["/etc/init.d/ucsschool-id-connector", "start"])

    end = datetime.now()
    duration = (end - start).total_seconds()
    logger.info("DONE improved first time sync. Processed in: %r seconds", duration)
