# -*- coding: utf-8 -*-

# Copyright 2022-2023 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
import datetime
import subprocess
from typing import Any, Dict, List

import faker
import pytest
import pytest_asyncio

import ucsschool_id_connector.plugin_loader
from ucsschool.kelvin.client import (
    NoObject,
    School as KelvinSchool,
    SchoolClass as KelvinSchoolClass,
    SchoolClassResource as KelvinSchoolClassResource,
    SchoolResource as KelvinSchoolResource,
    Session,
    User as KelvinUser,
    UserResource as KelvinUserResource,
)
from ucsschool_id_connector.ldap_access import LDAPAccess
from ucsschool_id_connector.models import SchoolAuthorityConfiguration

# load ID Broker plugin
ucsschool_id_connector.plugin_loader.load_plugins()
id_broker = pytest.importorskip("idbroker")
pytestmark = pytest.mark.id_broker

import idbroker.initial_sync as InitSync  # isort:skip # noqa: E402
from idbroker.id_broker_client import (  # isort:skip  # noqa: E402
    IDBrokerSchool,
    IDBrokerSchoolClass,
    IDBrokerUser,
)

fake = faker.Faker()
SYNC_TIMEOUT = 60

# TODO cleanup schools has to be done both on the ID-Connector & the ID Broker side.
# Users & classes are cleaned up implicitly when they are deleted on
# the ID Connector side.


@pytest_asyncio.fixture()
async def make_sender_user(
    kelvin_user_object,
    kelvin_session,
    kelvin_session_kwargs,
    docker_hostname,
    id_broker_fqdn,
):
    """
    Fixture factory to create users on the apps host system. They are created
    via the Kelvin-API and removed on both sides when the fixture goes out of scope.
    """
    created_users: List[Dict[str, Any]] = []
    _s_a_name: List[str] = []

    async def _delete_user(username: str, host: str) -> None:
        print(f"Deleting user {username!r} from host {host!r}...")
        # Creating a new session per user deletion, because of a problem in httpx/httpcore/h11.
        # I think it has a problem with the disconnect happening by Kelvin with FastAPI delete():
        # h11._util.LocalProtocolError: Too much data for declared Content-Length
        async with Session(**kelvin_session_kwargs(host)) as session:
            try:
                user = await KelvinUserResource(session=session).get(name=username)
                await user.delete()
                print(f"Success deleting user {username!r} from host {host!r}.")
            except NoObject:
                print(f"No user {username!r} on {host!r}.")

    async def _make_sender_user(roles=("student",), ous=("DEMOSCHOOL",), s_a_name="Traeger1"):
        """
        Creates a user on the hosts UCS system via Kelvin-API-Client

        :param roles: The new users roles
        :param ous: The new users ous
        :return: The json used to create the user via the API
        """
        nonlocal _s_a_name
        _s_a_name = [s_a_name]
        user_obj: KelvinUser = kelvin_user_object(roles, ous, session=kelvin_session(docker_hostname))
        password = user_obj.password
        await user_obj.save()
        user_obj.password = password
        user_obj_as_dict = user_obj.as_dict()
        created_users.append(user_obj_as_dict)
        print("Created new User in source system: {!r}".format(user_obj_as_dict))
        return user_obj_as_dict

    yield _make_sender_user

    for user_dict in created_users:
        await _delete_user(user_dict["name"], docker_hostname)
        # if the ID Connector works, there should be nothing to clean up.
        await _delete_user(f"{_s_a_name[0]}-{user_dict['name']}", id_broker_fqdn)


@pytest.mark.asyncio
async def test_school_exists_after_user_with_non_existing_school_is_synced(
    kelvin_school_on_sender,
    wait_for_kelvin_object_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    id_broker_school_auth_conf,
    make_school_authority,
    make_sender_user,
):
    """
    create kelvin user with non existing school
    -> school should be created on id-broker side
    -> the display_name should be set correct
    """
    school_authority = await make_school_authority(plugin_name="id_broker", **id_broker_school_auth_conf)
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])
    await asyncio.sleep(4)
    s_a_name = school_authority.name
    sender_user: Dict[str, Any] = await make_sender_user(
        ous=[kelvin_school_on_sender.name], s_a_name=s_a_name
    )
    kelvin_school_on_id_broker: KelvinSchool = await wait_for_kelvin_object_exists(
        resource_cls=KelvinSchoolResource,
        method="get",
        session=kelvin_session(id_broker_fqdn),
        name=f"{s_a_name}-{sender_user['school']}",
    )
    assert kelvin_school_on_id_broker.display_name == kelvin_school_on_sender.display_name


# IDBrokerPerSAGroupDispatcher


@pytest.mark.asyncio
async def test_school_exists_after_school_class_with_non_existin_school_is_synced(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
):
    """
    create kelvin school class with non existing school
    -> school should be created on id-broker side
    -> the display_name should be set correct
    """
    school_authority = await make_school_authority(plugin_name="id_broker", **id_broker_school_auth_conf)
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])
    s_a_name = school_authority.name
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
    )
    kelvin_school_on_id_broker: KelvinSchool = await wait_for_kelvin_object_exists(
        resource_cls=KelvinSchoolResource,
        method="get",
        session=kelvin_session(id_broker_fqdn),
        name=f"{s_a_name}-{sender_kelvin_school_class.school}",
    )
    assert kelvin_school_on_id_broker.display_name == kelvin_school_on_sender.display_name


@pytest.mark.asyncio
async def test_school_class_handle_attr_description_empty(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
):
    """
    Creating a school class with an empty descritption should not
    lead to an error but set it to an empty string on the id-broker side.
    """
    school_authority = await make_school_authority(plugin_name="id_broker", **id_broker_school_auth_conf)
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])
    s_a_name = school_authority.name
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
    )
    sender_kelvin_school_class.description = None
    await sender_kelvin_school_class.save()
    kelvin_school_class_on_id_broker: KelvinSchoolClass = await wait_for_kelvin_object_exists(
        resource_cls=KelvinSchoolClassResource,
        method="get",
        session=kelvin_session(id_broker_fqdn),
        name=sender_kelvin_school_class.name,
        school=f"{s_a_name}-{kelvin_school_on_sender.name}",
    )
    assert kelvin_school_class_on_id_broker.description == sender_kelvin_school_class.description


@pytest.mark.asyncio
async def test_handle_attr_users(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
):
    """
    Tests if the users of the kelvin class are placed inside on the
    id-broker side, see  IDBrokerPerSAGroupDispatcher._handle_attr_users

    This also tests implicitly
    - IDBrokerPerSAGroupDispatcher._handle_attr_name
    - IDBrokerPerSAGroupDispatcher._handle_attr_school
    """
    school_authority = await make_school_authority(plugin_name="id_broker", **id_broker_school_auth_conf)
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])
    s_a_name = school_authority.name
    sender_user: Dict[str, Any] = await make_sender_user(
        ous=[kelvin_school_on_sender.name], s_a_name=s_a_name
    )
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
        users=[sender_user["name"]],
    )
    # The class is 1st created without users, then the user is created and added to the class.
    # We'll have to wait with our check until then.
    timeout = datetime.datetime.now() + datetime.timedelta(seconds=SYNC_TIMEOUT)
    while datetime.datetime.now() < timeout:
        used_time = SYNC_TIMEOUT - (timeout - datetime.datetime.now()).seconds
        kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_object_exists(
            resource_cls=KelvinSchoolClassResource,
            method="get",
            session=kelvin_session(id_broker_fqdn),
            name=sender_kelvin_school_class.name,
            school=f"{s_a_name}-{kelvin_school_on_sender.name}",
        )
        try:
            assert kelvin_school_class.users == [f"{s_a_name}-{sender_user['name']}"]
            print(f"OK: After {used_time} sec members as expected.")
            break
        except AssertionError as exc:
            print(f"After {used_time} sec: {exc!r}")
        await asyncio.sleep(2)
    else:
        raise AssertionError("Members not as expected.")
    # test IDBrokerPerSAUserDispatcher._handle_attr_context
    remote_user: KelvinUser = await wait_for_kelvin_object_exists(
        resource_cls=KelvinUserResource,
        method="get",
        session=kelvin_session(id_broker_fqdn),
        name=f"{s_a_name}-{sender_user['name']}",
        school=f"{s_a_name}-{kelvin_school_on_sender.name}",
    )
    assert (
        sender_kelvin_school_class.name
        in remote_user.school_classes[f"{s_a_name}-{kelvin_school_on_sender.name}"]
    )
    assert set(sender_user["roles"]) == set(remote_user.roles)


@pytest.mark.asyncio
async def test_school_class_is_created_with_missing_users(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    wait_for_kelvin_object_not_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
):
    """
    Create school class with a user existing on the ID Broker and one missing there.
    -> school class should be created on ID Broker side
    -> existing user (#2) should be in the school class, missing user (#1) is ignored
    """
    school_authority: SchoolAuthorityConfiguration = await make_school_authority(
        plugin_name="id_broker", **id_broker_school_auth_conf
    )
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])
    s_a_name = school_authority.name
    sender_user1: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    sender_user2: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    for sender_user in (sender_user1, sender_user2):
        await wait_for_kelvin_object_exists(
            resource_cls=KelvinUserResource,
            method="get",
            session=kelvin_session(id_broker_fqdn),
            name=f"{s_a_name}-{sender_user['name']}",
            school=f"{s_a_name}-{kelvin_school_on_sender.name}",
        )
    ldap_access = LDAPAccess()
    results = await ldap_access.search(
        filter_s=f"(&(objectClass=posixAccount)(uid={sender_user1['name']}))",
        attributes=["entryUUID"],
    )
    assert len(results) == 1
    sender_user1_entry_uuid = results[0]["entryUUID"].value
    id_broker_user = IDBrokerUser(school_authority, "id_broker")
    await id_broker_user.delete(sender_user1_entry_uuid)
    assert not await id_broker_user.exists(sender_user1_entry_uuid)
    await wait_for_kelvin_object_not_exists(
        resource_cls=KelvinUserResource,
        method="get",
        session=kelvin_session(id_broker_fqdn),
        name=f"{s_a_name}-{sender_user1['name']}",
        school=f"{s_a_name}-{kelvin_school_on_sender.name}",
    )
    print(
        f"*** User {sender_user1['name']} was deleted on ID Broker system, user {sender_user2['name']} "
        f"exists."
    )
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
        users=[sender_user1["name"], sender_user2["name"]],
    )
    await sender_kelvin_school_class.save()
    # _fix_school_class_members() will 1st create the school class without members and then add only the
    # existing users. We'll find the class without members 1st, so give it a bit of time...
    timeout = datetime.datetime.now() + datetime.timedelta(seconds=SYNC_TIMEOUT)
    while datetime.datetime.now() < timeout:
        kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_object_exists(
            resource_cls=KelvinSchoolClassResource,
            method="get",
            session=kelvin_session(id_broker_fqdn),
            name=sender_kelvin_school_class.name,
            school=f"{s_a_name}-{kelvin_school_on_sender.name}",
        )
        used_time = SYNC_TIMEOUT - (timeout - datetime.datetime.now()).seconds
        try:
            assert kelvin_school_class.users == [f"{s_a_name}-{sender_user2['name']}"]
            print(f"OK: After {used_time} sec members as expected.")
            break
        except AssertionError as exc:
            print(f"After {used_time} sec: {exc!r}")
        await asyncio.sleep(2)
    else:
        raise AssertionError("Members not as expected.")


@pytest.mark.asyncio
async def test_school_class_is_updated_with_missing_users(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    wait_for_kelvin_object_not_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
):
    school_authority: SchoolAuthorityConfiguration = await make_school_authority(
        plugin_name="id_broker", **id_broker_school_auth_conf
    )
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])
    s_a_name = school_authority.name
    sender_user1: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    sender_user2: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    for sender_user in (sender_user1, sender_user2):
        await wait_for_kelvin_object_exists(
            resource_cls=KelvinUserResource,
            method="get",
            session=kelvin_session(id_broker_fqdn),
            name=f"{s_a_name}-{sender_user['name']}",
            school=f"{s_a_name}-{kelvin_school_on_sender.name}",
        )
    ldap_access = LDAPAccess()
    results = await ldap_access.search(
        filter_s=f"(&(objectClass=posixAccount)(uid={sender_user1['name']}))",
        attributes=["entryUUID"],
    )
    assert len(results) == 1
    sender_user1_entry_uuid = results[0]["entryUUID"].value
    id_broker_user = IDBrokerUser(school_authority, "id_broker")
    await id_broker_user.delete(sender_user1_entry_uuid)
    assert not await id_broker_user.exists(sender_user1_entry_uuid)
    await wait_for_kelvin_object_not_exists(
        resource_cls=KelvinUserResource,
        method="get",
        session=kelvin_session(id_broker_fqdn),
        name=f"{s_a_name}-{sender_user1['name']}",
        school=f"{s_a_name}-{kelvin_school_on_sender.name}",
    )
    print(
        f"*** User {sender_user1['name']} was deleted on ID Broker system, user {sender_user2['name']} "
        f"exists."
    )
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
        users=[],
    )
    await sender_kelvin_school_class.save()
    kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_object_exists(
        resource_cls=KelvinSchoolClassResource,
        method="get",
        session=kelvin_session(id_broker_fqdn),
        name=sender_kelvin_school_class.name,
        school=f"{s_a_name}-{kelvin_school_on_sender.name}",
    )
    assert kelvin_school_class.users == []

    sender_kelvin_school_class.users.extend([sender_user1["name"], sender_user2["name"]])
    await sender_kelvin_school_class.save()
    expected = [f"{s_a_name}-{sender_user2['name']}"]
    # If we check immediately, we'll be faster than the ID Connector and will find an empty class.
    timeout = datetime.datetime.now() + datetime.timedelta(seconds=SYNC_TIMEOUT)
    while datetime.datetime.now() < timeout:
        kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_object_exists(
            resource_cls=KelvinSchoolClassResource,
            method="get",
            session=kelvin_session(id_broker_fqdn),
            name=sender_kelvin_school_class.name,
            school=f"{s_a_name}-{kelvin_school_on_sender.name}",
        )
        found = kelvin_school_class.users
        used_time = SYNC_TIMEOUT - (timeout - datetime.datetime.now()).seconds
        try:
            assert found == expected
            print(f"OK: After {used_time} sec members as expected.")
            break
        except AssertionError as exc:
            print(f"After {used_time} sec: {exc!r}")
        await asyncio.sleep(2)
    else:
        raise AssertionError("Members not as expected.")


def run_inital_sync_script(call_arguments: list, conf_name: str) -> None:
    import os

    os.system(
        "chmod ugo+rwx /var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/plugins/packages/idbroker/initial_sync.py"  # noqa: E501
    )

    call_pipe = "/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/plugins/packages/idbroker/initial_sync.py"  # noqa: E501
    call_pipe += " -c /var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/school_authorities"
    for call_arg in call_arguments:
        call_pipe += f" {call_arg}"

    os.system(call_pipe)


@pytest.mark.asyncio
async def test_initial_sync_dry_run(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    wait_for_kelvin_object_not_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
    mock_env,
):
    # setup
    school_authority: SchoolAuthorityConfiguration = await make_school_authority(
        plugin_name="id_broker", **id_broker_school_auth_conf
    )
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])

    SCHOOL_CLIENT = IDBrokerSchool(school_authority, "id_broker")
    CLASSES_CLIENT = IDBrokerSchoolClass(school_authority, "id_broker")

    ### initial status tracking
    initial_ous = await InitSync.get_ous("*")
    initial_classes = await InitSync.get_classes("*")

    initial_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in initial_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    initial_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in initial_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]
    for initial_remote in initial_ous_remote:
        if not any(initial_remote == initial_local for initial_local in initial_ous):
            raise AssertionError(
                f"Values of OU {initial_remote.name} differ (local vs remote)\n{initial_remote.dict()}"
            )

    for initial_remote in initial_classes_remote:
        if not any(initial_remote == initial_local for initial_local in initial_classes_remote):
            raise AssertionError(
                f"Values of CLASS {initial_remote.name} differ (local vs remote) \n{initial_remote.dict()}"
            )

    ### setup differences
    s_a_name = school_authority.name
    kelvin_users = []
    for i in range(0, 2):
        sender_user: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
        kelvin_users.append(sender_user)
    for user in kelvin_users:
        await wait_for_kelvin_object_exists(
            resource_cls=KelvinUserResource,
            method="get",
            session=kelvin_session(id_broker_fqdn),
            name=f"{s_a_name}-{user['name']}",
            school=f"{s_a_name}-{kelvin_school_on_sender.name}",
        )
    ### run script in dry_mode
    run_inital_sync_script(call_arguments=["-d -t 9"], conf_name=school_authority.name)

    ### post script check

    post_ous = initial_ous + await InitSync.get_ous(kelvin_school_on_sender.name)
    post_classes = initial_classes + await InitSync.get_classes(kelvin_school_on_sender.name)

    post_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in post_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    post_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in post_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]
    for post_remote in post_ous_remote:
        for initial_remote in initial_ous_remote:
            if post_remote.id == initial_remote.id:
                if not post_remote == initial_remote:
                    raise AssertionError(
                        f"Values of OU {post_remote.name} differ after initial sync\n"
                        f"initial={initial_remote.dict()}\npost=\n{post_remote.dict()}"
                    )

    for post_remote in post_classes_remote:
        for initial_remote in initial_classes_remote:
            if post_remote.id == initial_remote.id:
                if not post_remote == initial_remote:
                    raise AssertionError(
                        f"Values of CLASSES {post_remote.name} differ after initial sync\n"
                        f"initial={initial_remote.dict()}\npost=\n{post_remote.dict()}"
                    )


@pytest.mark.asyncio
async def test_initial_sync_single_ou(
    wait_for_kelvin_object_exists,
    wait_for_kelvin_object_not_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    kelvin_school_on_sender,
    make_sender_user,
    make_kelvin_school_class_on_id_connector,
    create_schools,
    mock_env,
):
    school_authority: SchoolAuthorityConfiguration = await make_school_authority(
        plugin_name="id_broker", **id_broker_school_auth_conf
    )
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])

    SCHOOL_CLIENT = IDBrokerSchool(school_authority, "id_broker")
    CLASSES_CLIENT = IDBrokerSchoolClass(school_authority, "id_broker")

    s_a_name = school_authority.name

    ### setup differences

    kelvin_users = []
    for i in range(0, 2):
        sender_user: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
        kelvin_users.append(sender_user)
    for user in kelvin_users:
        await wait_for_kelvin_object_exists(
            resource_cls=KelvinUserResource,
            method="get",
            session=kelvin_session(id_broker_fqdn),
            name=f"{s_a_name}-{user['name']}",
            school=f"{s_a_name}-{kelvin_school_on_sender.name}",
        )

    ### initial status tracking
    initial_ous = await InitSync.get_ous("*")
    initial_classes = await InitSync.get_classes("*")

    initial_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in initial_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    initial_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in initial_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]

    for initial_remote in initial_ous_remote:
        if not any(initial_remote == initial_local for initial_local in initial_ous):
            raise AssertionError(
                f"Values of OU {initial_remote.name} differ (local vs remote)\n{initial_remote.dict()}"
            )

    for initial_remote in initial_classes_remote:
        if not any(initial_remote == initial_local for initial_local in initial_classes_remote):
            raise AssertionError(
                f"Values of CLASS {initial_remote.name} differ (local vs remote) \n{initial_remote.dict()}"
            )

    ### run script in single mode
    run_inital_sync_script(
        call_arguments=[f"-m -t 9 -s {kelvin_school_on_sender.name}"],
        conf_name=school_authority.name,
    )

    ### post script check

    post_ous = initial_ous + await InitSync.get_ous(kelvin_school_on_sender.name)
    post_classes = initial_classes + await InitSync.get_classes(kelvin_school_on_sender.name)

    post_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in post_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    post_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in post_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]
    for post_remote in post_ous_remote:
        if not any(post_remote == post_local for post_local in post_ous):
            raise AssertionError(
                f"Values of OU {post_remote.name} differ after initial sync\n{post_remote.dict()}"
            )

    for post_remote in post_classes_remote:
        similar = False
        for post_local in post_classes:
            if post_local == post_remote:
                similar = True

        if not similar:
            print(post_remote)
            print(post_classes)
            raise AssertionError(
                f"Values of CLASS {post_remote.name} differ after initial sync\n{post_remote.dict()}"
            )


@pytest.mark.asyncio
async def test_initial_sync_only_populated_ous_and_classes(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    wait_for_kelvin_object_not_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
    mock_env,
):
    # setup
    school_authority: SchoolAuthorityConfiguration = await make_school_authority(
        plugin_name="id_broker", **id_broker_school_auth_conf
    )
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])

    SCHOOL_CLIENT = IDBrokerSchool(school_authority, "id_broker")
    CLASSES_CLIENT = IDBrokerSchoolClass(school_authority, "id_broker")

    run_inital_sync_script(call_arguments=["-t 9"], conf_name=school_authority.name)

    ### initial status tracking
    initial_ous = await InitSync.get_ous("*")
    initial_classes = await InitSync.get_classes("*")

    initial_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in initial_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    initial_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in initial_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]

    for initial_remote in initial_ous_remote:
        if not any(initial_remote == initial_local for initial_local in initial_ous):
            raise AssertionError(
                f"Values of OU {initial_remote.name} differ (local vs remote)\n{initial_remote.dict()}"
            )

    for initial_remote in initial_classes_remote:
        if not any(initial_remote == initial_local for initial_local in initial_classes_remote):
            raise AssertionError(
                f"Values of CLASS {initial_remote.name} differ (local vs remote) \n{initial_remote.dict()}"
            )

    return


@pytest.mark.asyncio
async def test_initial_sync_with_members(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    wait_for_kelvin_object_not_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
    mock_env,
):
    school_authority: SchoolAuthorityConfiguration = await make_school_authority(
        plugin_name="id_broker", **id_broker_school_auth_conf
    )
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])

    SCHOOL_CLIENT = IDBrokerSchool(school_authority, "id_broker")
    CLASSES_CLIENT = IDBrokerSchoolClass(school_authority, "id_broker")

    ### initial status tracking
    initial_ous = await InitSync.get_ous("*")
    initial_classes = await InitSync.get_classes("*")

    initial_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in initial_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    initial_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in initial_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]

    for initial_remote in initial_ous_remote:
        if not any(initial_remote == initial_local for initial_local in initial_ous):
            raise AssertionError(
                f"Values of OU {initial_remote.name} differ (local vs remote)\n{initial_remote.dict()}"
            )

    for initial_remote in initial_classes_remote:
        if not any(initial_remote == initial_local for initial_local in initial_classes_remote):
            raise AssertionError(
                f"Values of CLASS {initial_remote.name} differ (local vs remote) \n{initial_remote.dict()}"
            )

    ### run script in single mode
    run_inital_sync_script(call_arguments=["-t 9 -m"], conf_name=school_authority.name)
    post_classes = initial_classes + await InitSync.get_classes(kelvin_school_on_sender.name)

    post_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in post_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]

    for post_remote in post_classes_remote:
        assert post_remote != []
        # if not post_remote.members:
        #    raise AssertionError(f"Values of CLASS {post_remote.name}
        #    differ after initial sync\n{post_remote.dict()}")

    return


@pytest.mark.asyncio
async def test_initial_sync_without_members(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_object_exists,
    wait_for_kelvin_object_not_exists,
    kelvin_session,
    id_connector_host_name,
    id_broker_fqdn,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
    mock_env,
):
    school_authority: SchoolAuthorityConfiguration = await make_school_authority(
        plugin_name="id_broker", **id_broker_school_auth_conf
    )
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])

    SCHOOL_CLIENT = IDBrokerSchool(school_authority, "id_broker")
    CLASSES_CLIENT = IDBrokerSchoolClass(school_authority, "id_broker")

    ### initial status tracking
    initial_ous = await InitSync.get_ous("*")
    initial_classes = await InitSync.get_classes("*")

    initial_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in initial_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    initial_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in initial_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]

    for initial_remote in initial_ous_remote:
        if not any(initial_remote == initial_local for initial_local in initial_ous):
            raise AssertionError(
                f"Values of OU {initial_remote.name} differ (local vs remote)\n{initial_remote.dict()}"
            )

    for initial_remote in initial_classes_remote:
        if not any(initial_remote == initial_local for initial_local in initial_classes_remote):
            raise AssertionError(
                f"Values of CLASS {initial_remote.name} differ (local vs remote) \n{initial_remote.dict()}"
            )

    ### run script in single mode
    run_inital_sync_script(call_arguments=["-t 9"], conf_name=school_authority.name)

    post_ous = initial_ous + await InitSync.get_ous(kelvin_school_on_sender.name)
    post_classes = initial_classes + await InitSync.get_classes(kelvin_school_on_sender.name)

    post_ous_remote = [
        await SCHOOL_CLIENT.get(ou.id) for ou in post_ous if await SCHOOL_CLIENT.exists(ou.id) is True
    ]
    post_classes_remote = [
        await CLASSES_CLIENT.get(school_class.id)
        for school_class in post_classes
        if await CLASSES_CLIENT.exists(school_class.id) is True
    ]
    for post_remote in post_ous_remote:
        if not any(post_remote == post_local for post_local in post_ous):
            raise AssertionError(
                f"Values of OU {post_remote.name} differ after initial sync\n{post_remote.dict()}"
            )

    for post_remote in post_classes_remote:
        assert post_remote.members == []
        # if not post_remote.members:
        #    raise AssertionError(f"Values of CLASS {post_remote.name}
        #    differ after initial sync\n{post_remote.dict()}")
