# -*- coding: utf-8 -*-

# Copyright 2022-2023 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
import datetime
import subprocess
from typing import Any, Dict, List

import faker
import pytest
import pytest_asyncio

import ucsschool_id_connector.plugin_loader
from ucsschool.kelvin.client import (
    NoObject,
    School as KelvinSchool,
    SchoolClass as KelvinSchoolClass,
    Session,
    User as KelvinUser,
    UserResource as KelvinUserResource,
)
from ucsschool_id_connector.ldap_access import LDAPAccess

# load ID Broker plugin
ucsschool_id_connector.plugin_loader.load_plugins()
id_broker = pytest.importorskip("idbroker")
pytestmark = pytest.mark.id_broker

from idbroker.id_broker_client import (  # isort:skip  # noqa: E402
    IDBrokerUser,
)

fake = faker.Faker()
SYNC_TIMEOUT = 60

# TODO cleanup schools has to be done both on the ID-Connector & the ID Broker side.
# Users & classes are cleaned up implicitly when they are deleted on
# the ID Connector side.


@pytest_asyncio.fixture()
async def make_sender_user(
    kelvin_user_object,
    kelvin_session,
    kelvin_session_kwargs,
    docker_hostname,
    id_broker_fqdn,
):
    """
    Fixture factory to create users on the apps host system. They are created
    via the Kelvin-API and removed on both sides when the fixture goes out of scope.
    """
    created_users: List[Dict[str, Any]] = []
    _s_a_name: List[str] = []

    async def _delete_user(username: str, host: str) -> None:
        print(f"Deleting user {username!r} from host {host!r}...")
        # Creating a new session per user deletion, because of a problem in httpx/httpcore/h11.
        # I think it has a problem with the disconnect happening by Kelvin with FastAPI delete():
        # h11._util.LocalProtocolError: Too much data for declared Content-Length
        async with Session(**kelvin_session_kwargs(host)) as session:
            try:
                user = await KelvinUserResource(session=session).get(name=username)
                await user.delete()
                print(f"Success deleting user {username!r} from host {host!r}.")
            except NoObject:
                print(f"No user {username!r} on {host!r}.")

    async def _make_sender_user(roles=("student",), ous=("DEMOSCHOOL",), s_a_name="Traeger1"):
        """
        Creates a user on the hosts UCS system via Kelvin-API-Client

        :param roles: The new users roles
        :param ous: The new users ous
        :return: The json used to create the user via the API
        """
        nonlocal _s_a_name
        _s_a_name = [s_a_name]
        user_obj: KelvinUser = kelvin_user_object(roles, ous, session=kelvin_session(docker_hostname))
        password = user_obj.password
        await user_obj.save()
        user_obj.password = password
        user_obj_as_dict = user_obj.as_dict()
        created_users.append(user_obj_as_dict)
        print("Created new User in source system: {!r}".format(user_obj_as_dict))
        return user_obj_as_dict

    yield _make_sender_user

    for user_dict in created_users:
        await _delete_user(user_dict["name"], docker_hostname)
        # if the ID Connector works, there should be nothing to clean up.
        await _delete_user(f"{_s_a_name[0]}-{user_dict['name']}", id_broker_fqdn)


async def set_school_authority_on_id_connector(id_broker_school_auth_conf, make_school_authority):
    school_authority = await make_school_authority(plugin_name="id_broker", **id_broker_school_auth_conf)
    subprocess.Popen(["/etc/init.d/ucsschool-id-connector", "restart"])
    return school_authority


@pytest.mark.asyncio
async def test_school_exists_after_user_with_non_existing_school_is_synced(
    kelvin_school_on_sender,
    wait_for_kelvin_school_on_id_broker,
    id_broker_school_auth_conf,
    make_school_authority,
    make_sender_user,
):
    """
    create kelvin user with non-existing school
    -> school should be created on id broker side
    -> the display_name should be set correct
    """
    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    s_a_name = school_authority.name
    sender_user: Dict[str, Any] = await make_sender_user(
        ous=[kelvin_school_on_sender.name], s_a_name=s_a_name
    )
    kelvin_school_on_id_broker: KelvinSchool = await wait_for_kelvin_school_on_id_broker(
        s_a_name=s_a_name, name=sender_user["school"]
    )
    assert kelvin_school_on_id_broker.display_name == kelvin_school_on_sender.display_name


@pytest.mark.parametrize("initial_import_mode", ["unset", True, False])
@pytest.mark.parametrize("role", ["student", "teacher"])
@pytest.mark.asyncio
async def test_initial_import_mode_user(
    kelvin_schools_on_sender,
    wait_for_kelvin_user_on_id_broker,
    id_broker_school_auth_conf,
    make_school_authority,
    make_sender_user,
    initial_import_mode,
    role,
):
    """
    set initial import mode  -> user should be created without classes
    unset + disable initial import mode -> user should be created with classes
    """
    if initial_import_mode != "unset":
        id_broker_school_auth_conf["plugin_configs"]["id_broker"][
            "initial_import_mode"
        ] = initial_import_mode
    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    sender_school_1, sender_school2 = await kelvin_schools_on_sender(2)
    s_a_name = school_authority.name
    sender_user: Dict[str, Any] = await make_sender_user(
        ous=[sender_school_1.name, sender_school2.name], s_a_name=s_a_name, roles=(role,)
    )
    remote_user: KelvinUser = await wait_for_kelvin_user_on_id_broker(
        s_a_name=s_a_name, school=sender_school_1.name, name=sender_user["name"]
    )
    if initial_import_mode == 1:
        assert remote_user.school_classes == {}
    else:
        for key, value in sender_user["school_classes"].items():
            assert set(value) == set(remote_user.school_classes[f"{s_a_name}-{key}"])


# IDBrokerPerSAGroupDispatcher


@pytest.mark.parametrize("initial_import_mode", ["unset", True, False])
@pytest.mark.asyncio
async def test_initial_import_mode_school_class(
    kelvin_school_on_sender,
    wait_for_kelvin_school_class_on_id_broker,
    id_broker_school_auth_conf,
    make_school_authority,
    make_kelvin_school_class_on_id_connector,
    initial_import_mode,
):
    """
    set initial import mode -> should not create school_class
    unset + disable initial import mode -> should create school_class
    """
    if initial_import_mode != "unset":
        id_broker_school_auth_conf["plugin_configs"]["id_broker"][
            "initial_import_mode"
        ] = initial_import_mode
    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    s_a_name = school_authority.name
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
    )
    try:
        await wait_for_kelvin_school_class_on_id_broker(
            s_a_name=s_a_name,
            school=kelvin_school_on_sender.name,
            name=sender_kelvin_school_class.name,
        )
    except AssertionError:
        if initial_import_mode != 1:
            raise AssertionError(
                f"Class should have been created in normal mode (initial_import_mode={initial_import_mode})"
            )
    # in initial_import_mode the class should not be created


@pytest.mark.asyncio
async def test_school_exists_after_school_class_with_non_existing_school_is_synced(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_school_on_id_broker,
    make_school_authority,
    id_broker_school_auth_conf,
):
    """
    create kelvin school class with non-existing school
    -> school should be created on id broker side
    -> the display_name should be set correct
    """
    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    s_a_name = school_authority.name
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
    )
    kelvin_school_on_id_broker: KelvinSchool = await wait_for_kelvin_school_on_id_broker(
        s_a_name=s_a_name, name=sender_kelvin_school_class.school
    )
    assert kelvin_school_on_id_broker.display_name == kelvin_school_on_sender.display_name


@pytest.mark.asyncio
async def test_school_class_handle_attr_description_empty(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_school_class_on_id_broker,
    make_school_authority,
    id_broker_school_auth_conf,
):
    """
    Creating a school class with an empty description should not
    lead to an error but set it to an empty string on the id broker side.
    """

    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    s_a_name = school_authority.name
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
    )
    sender_kelvin_school_class.description = None
    await sender_kelvin_school_class.save()
    kelvin_school_class_on_id_broker: KelvinSchoolClass = (
        await wait_for_kelvin_school_class_on_id_broker(
            s_a_name=s_a_name,
            school=kelvin_school_on_sender.name,
            name=sender_kelvin_school_class.name,
        )
    )
    assert kelvin_school_class_on_id_broker.description == sender_kelvin_school_class.description


@pytest.mark.asyncio
async def test_handle_attr_users(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_school_class_on_id_broker,
    wait_for_kelvin_user_on_id_broker,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
):
    """
    Tests if the users of the kelvin class are placed inside on the
    id broker side, see  IDBrokerPerSAGroupDispatcher._handle_attr_users

    This also tests implicitly
    - IDBrokerPerSAGroupDispatcher._handle_attr_name
    - IDBrokerPerSAGroupDispatcher._handle_attr_school
    """
    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    s_a_name = school_authority.name
    sender_user: Dict[str, Any] = await make_sender_user(
        ous=[kelvin_school_on_sender.name], s_a_name=s_a_name
    )
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
        users=[sender_user["name"]],
    )
    # The class is 1st created without users, then the user is created and added to the class.
    # We'll have to wait with our check until then.
    timeout = datetime.datetime.now() + datetime.timedelta(seconds=SYNC_TIMEOUT)
    while datetime.datetime.now() < timeout:
        used_time = SYNC_TIMEOUT - (timeout - datetime.datetime.now()).seconds
        kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_school_class_on_id_broker(
            s_a_name=s_a_name,
            school=kelvin_school_on_sender.name,
            name=sender_kelvin_school_class.name,
        )
        try:
            assert kelvin_school_class.users == [f"{s_a_name}-{sender_user['name']}"]
            print(f"OK: After {used_time} sec members as expected.")
            break
        except AssertionError as exc:
            print(f"After {used_time} sec: {exc!r}")
        await asyncio.sleep(2)
    else:
        raise AssertionError("Members not as expected.")
    # test IDBrokerPerSAUserDispatcher._handle_attr_context
    remote_user: KelvinUser = await wait_for_kelvin_user_on_id_broker(
        s_a_name=s_a_name, name=sender_user["name"], school=kelvin_school_on_sender.name
    )
    assert (
        sender_kelvin_school_class.name
        in remote_user.school_classes[f"{s_a_name}-{kelvin_school_on_sender.name}"]
    )
    assert set(sender_user["roles"]) == set(remote_user.roles)


@pytest.mark.asyncio
async def test_school_class_is_created_with_missing_users(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_user_on_id_broker,
    wait_for_kelvin_school_class_on_id_broker,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
):
    """
    Create school class with a user existing on the ID Broker and one missing there.
    -> school class should be created on ID Broker side
    -> existing user (#2) should be in the school class, missing user (#1) is ignored
    """
    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    s_a_name = school_authority.name
    sender_user1: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    sender_user2: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    for sender_user in (sender_user1, sender_user2):
        await wait_for_kelvin_user_on_id_broker(
            s_a_name=s_a_name, name=sender_user["name"], school=kelvin_school_on_sender.name
        )
    ldap_access = LDAPAccess()
    results = await ldap_access.search(
        filter_s=f"(&(objectClass=posixAccount)(uid={sender_user1['name']}))",
        attributes=["entryUUID"],
    )
    assert len(results) == 1
    sender_user1_entry_uuid = results[0]["entryUUID"].value
    id_broker_user = IDBrokerUser(school_authority, "id_broker")
    await id_broker_user.delete(sender_user1_entry_uuid)
    assert not await id_broker_user.exists(sender_user1_entry_uuid)
    await wait_for_kelvin_user_on_id_broker(
        s_a_name=s_a_name,
        name=sender_user1["name"],
        school=kelvin_school_on_sender.name,
        must_exist=False,
    )
    print(
        f"*** User {sender_user1['name']} was deleted on ID Broker system, user {sender_user2['name']} "
        f"exists."
    )
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
        users=[sender_user1["name"], sender_user2["name"]],
    )
    await sender_kelvin_school_class.save()
    # _fix_school_class_members() will 1st create the school class without members and then add only the
    # existing users. We'll find the class without members 1st, so give it a bit of time...
    timeout = datetime.datetime.now() + datetime.timedelta(seconds=SYNC_TIMEOUT)
    while datetime.datetime.now() < timeout:
        kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_school_class_on_id_broker(
            s_a_name=s_a_name, name=sender_kelvin_school_class.name, school=kelvin_school_on_sender.name
        )
        used_time = SYNC_TIMEOUT - (timeout - datetime.datetime.now()).seconds
        try:
            assert kelvin_school_class.users == [f"{s_a_name}-{sender_user2['name']}"]
            print(f"OK: After {used_time} sec members as expected.")
            break
        except AssertionError as exc:
            print(f"After {used_time} sec: {exc!r}")
        await asyncio.sleep(2)
    else:
        raise AssertionError("Members not as expected.")


@pytest.mark.asyncio
async def test_school_class_is_updated_with_missing_users(
    kelvin_school_on_sender,
    make_kelvin_school_class_on_id_connector,
    wait_for_kelvin_user_on_id_broker,
    wait_for_kelvin_school_class_on_id_broker,
    make_school_authority,
    id_broker_school_auth_conf,
    make_sender_user,
):
    """
    Create school class with NO users, then add two users, from which only one exists on the ID Broker.
    -> school class should be updated on ID Broker side
    -> existing user should be in the school class, missing user is ignored
    """
    school_authority = await set_school_authority_on_id_connector(
        id_broker_school_auth_conf, make_school_authority
    )
    s_a_name = school_authority.name
    sender_user1: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    sender_user2: Dict[str, Any] = await make_sender_user(ous=[kelvin_school_on_sender.name])
    for sender_user in (sender_user1, sender_user2):
        await wait_for_kelvin_user_on_id_broker(
            s_a_name=s_a_name, name=sender_user["name"], school=kelvin_school_on_sender.name
        )
    ldap_access = LDAPAccess()
    results = await ldap_access.search(
        filter_s=f"(&(objectClass=posixAccount)(uid={sender_user1['name']}))",
        attributes=["entryUUID"],
    )
    assert len(results) == 1
    sender_user1_entry_uuid = results[0]["entryUUID"].value
    id_broker_user = IDBrokerUser(school_authority, "id_broker")
    await id_broker_user.delete(sender_user1_entry_uuid)
    assert not await id_broker_user.exists(sender_user1_entry_uuid)
    await wait_for_kelvin_user_on_id_broker(
        s_a_name=s_a_name,
        name=sender_user1["name"],
        school=kelvin_school_on_sender.name,
        must_exist=False,
    )
    print(
        f"*** User {sender_user1['name']} was deleted on ID Broker system, user {sender_user2['name']} "
        f"exists."
    )
    sender_kelvin_school_class: KelvinSchoolClass = await make_kelvin_school_class_on_id_connector(
        school_name=kelvin_school_on_sender.name,
        users=[],
    )
    await sender_kelvin_school_class.save()
    kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_school_class_on_id_broker(
        s_a_name=s_a_name, name=sender_kelvin_school_class.name, school=kelvin_school_on_sender.name
    )
    assert kelvin_school_class.users == []

    sender_kelvin_school_class.users.extend([sender_user1["name"], sender_user2["name"]])
    await sender_kelvin_school_class.save()
    expected = [f"{s_a_name}-{sender_user2['name']}"]
    # If we check immediately, we'll be faster than the ID Connector and will find an empty class.
    timeout = datetime.datetime.now() + datetime.timedelta(seconds=SYNC_TIMEOUT)
    while datetime.datetime.now() < timeout:
        kelvin_school_class: KelvinSchoolClass = await wait_for_kelvin_school_class_on_id_broker(
            s_a_name=s_a_name, name=sender_kelvin_school_class.name, school=kelvin_school_on_sender.name
        )
        found = kelvin_school_class.users
        used_time = SYNC_TIMEOUT - (timeout - datetime.datetime.now()).seconds
        try:
            assert found == expected
            print(f"OK: After {used_time} sec members as expected.")
            break
        except AssertionError as exc:
            print(f"After {used_time} sec: {exc!r}")
        await asyncio.sleep(2)
    else:
        raise AssertionError("Members not as expected.")
